﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Selections;
using STFXmlParser;
using CTNet;

namespace CalcHelper
{
    public partial class ToolBelt : Form
    {
        DocCalcs docCalcs;
        private DocFile _docFile;

        private CTClient _ctClient;

        public enum tbTabPage { boxSel, codeNav, fcnPick, snipPick };

        // One or more selections are implied if no 'Single' in text.
        public enum ctSelMode { none, crStdCalcSingleNonCal, crStdCalcCalNonCal }

        public tbTabPage tabPage;
        public ctSelMode selMode;

        // This holds selected boxes until we process them for an operation.
        public List<BoxSelect> boxSelBuff = new List<BoxSelect>();

        /*
         * ***********************
         * Code Navigator stuff:
         */
        public enum codeNavList { all, docFcns, comMrkrs, orphBoxes };
        public int codeNavFcnListIndex;
        public int codeNavComListIndex;
        public int codeNavOrphanListIndex;

        /*
         * ***********************
         */

        public ToolBelt(ref CTClient ctClient, ref DocFile docFile, ref DocCalcs dc, tbTabPage tp)
        {
            InitializeComponent();

            _ctClient = ctClient;
            docCalcs = dc;
            _docFile = docFile;

            TabPageSetup(tp);
            SelectTabPage(tp);

            // Subscribing to the Form's Box Selection event.
            Form1.BoxSel += Form1_CalcToolBoxSel;
            
            /*
             * ***********************
             * Code Navigator Initialization:
             */
            foreach (string item in docFile.ctrlArrays)
            {
                if (item.Contains("getCheckBoxList"))
                {
                    ctrlArrayList.Items.Add("Check Box Control");
                }
                else if (item.Contains("calc_controlled"))
                {
                    ctrlArrayList.Items.Add("Calc Controlled");
                }
            }

            // Turn off the Code Navigator's dynamic lists selections.
            CodeNavListSelectMode(codeNavList.all, false);

            // The Datasources are modified by SetCodeNavLists() in Pdf.cs:
            docFcnList.DataSource = docFile.codeNavDocFcns;
            comMarkerList.DataSource = docFile.codeNavComMarkers;
            orphanBoxList.DataSource = docFile.codeNavOrphanBoxes;

            // Turn on the Code Navigator's dynamic lists selections.
            CodeNavListSelectMode(codeNavList.all, true);

            codeNavFcnListIndex = -1;
            codeNavComListIndex = -1;
            codeNavOrphanListIndex = -1;

            /*
             * ***********************
             */
        }

        /*
         * Called from:
         * 
         * ToolBelt() constructor;
         * Pdf.cs: SetCodeNavLists();
         * 
         * mode - 'true' means to set the Code Navigator's dynamic list's
         *          selection mode to 'One'. 
         *  
         *        'false' means to set the selection mode to 'None'. This
         *          will prevent a list item from being automatically
         *          selected, namely, the first item.
         *  
         * These are the lists that can be dynamically added to or deleted
         * from as a result of a Refresh operation. Only the user makes a
         * selection, we do not want a selection made 'for them'. When a
         * selection is made we will 'goto' that code location in the
         * editor.
         */
        public void CodeNavListSelectMode(codeNavList listId, bool mode)
        {
            SelectionMode selMode = (mode ? SelectionMode.One : SelectionMode.None);

            switch (listId)
            {
                case (codeNavList.docFcns):

                    docFcnList.SelectionMode = selMode;

                    break;

                case (codeNavList.comMrkrs):

                    comMarkerList.SelectionMode = selMode;

                    break;

                case (codeNavList.orphBoxes):

                    orphanBoxList.SelectionMode = selMode;

                    break;

                default:

                    docFcnList.SelectionMode = selMode;
                    comMarkerList.SelectionMode = selMode;
                    orphanBoxList.SelectionMode = selMode;

                    break;
            }
        }

        /*
         * Called from:
         * 
         * Pdf.cs - SetCodeNavLists().
         * 
         * We try to reset the selections to what they were before
         * the refresh. We say 'try' because the lists may change
         * in length, in which case we reset the selections
         * accordingly.
         */
        public void SetCodeNavListSel(codeNavList listId)
        {
            switch (listId)
            {
                case (codeNavList.docFcns):
                    
                    if (codeNavFcnListIndex > docFcnList.Items.Count - 1)
                    {
                        codeNavFcnListIndex = docFcnList.Items.Count - 1;
                    }

                    docFcnList.SelectedIndex = codeNavFcnListIndex;

                    break;

                case (codeNavList.comMrkrs):

                    if (codeNavComListIndex > comMarkerList.Items.Count - 1)
                    {
                        codeNavComListIndex = comMarkerList.Items.Count - 1;
                    }

                    comMarkerList.SelectedIndex = codeNavComListIndex;

                    break;

                case (codeNavList.orphBoxes):

                    if (codeNavOrphanListIndex > orphanBoxList.Items.Count - 1)
                    {
                        codeNavOrphanListIndex = orphanBoxList.Items.Count - 1;
                    }

                    orphanBoxList.SelectedIndex = codeNavOrphanListIndex;

                    break;
            }
        }

        public void SelectTabPage(tbTabPage tp)
        {
            switch (tp)
            {
                case (tbTabPage.boxSel):

                    toolBeltTabCtrl.SelectedTab = boxSelTab;

                    break;

                case (tbTabPage.codeNav):

                    toolBeltTabCtrl.SelectedTab = codeNavTab;

                    break;

                case (tbTabPage.fcnPick):

                    toolBeltTabCtrl.SelectedTab = fcnPickTab;

                    break;

                case (tbTabPage.snipPick):

                    toolBeltTabCtrl.SelectedTab = snipPickTab;

                    break;
            }
        }

        /*
         * Our handler for the forms's box selection event.
        */
        public void Form1_CalcToolBoxSel(object sender, Form1.BoxSelArgs e)
        {
            if (e != null)
            {
                if (e.boxSel != null)
                {
                    if (e.boxSel.Count > 0)
                    {
                        // Store the boxes in the buffer...need to re-think this, some boxes may be
                        // unselected, etc..do not want duplicates...
                        // [11/20/18] I am thinking that the Parent form checks for this and stops it
                        // right there if this happens.
                        boxSelBuff.AddRange(e.boxSel);

                        // Determine the active tab page and process the selected box(es).
                        switch (tabPage)
                        {
                            case (tbTabPage.boxSel):

                                ProcessCreateStdBoxSel();

                                break;

                            case (tbTabPage.codeNav):

                                break;

                            case (tbTabPage.fcnPick):

                                ProcessEditViewBoxSel();

                                break;

                            case (tbTabPage.snipPick):

                                break;
                        }
                    }
                }
            }
        }

        private void TabPageSetup(tbTabPage tp)
        {
            tabPage = tp;

            selMode = ctSelMode.none;

            switch (tp)
            {
                case (tbTabPage.boxSel):

                    createStdFormatComboBox.SelectedIndex = 1;
                    createStdAttrComboBox.SelectedIndex = 0;
                    createStdBoxAttrComboBox.SelectedIndex = 0;

                    createStdCalcMsg1Label.Text = "Click the Start button to start creating a calculation.";
                    createStdCalcMsg2Label.Text = "Tip: Now is the time to set the Calc Format, and any other Calc/Box attributes if applicable.";

                    break;

                case (tbTabPage.codeNav):

                    break;

                case (tbTabPage.fcnPick):

                    break;

                case (tbTabPage.snipPick):

                    break;
            }
        }

        private void ProcessEditViewBoxSel()
        {
            /**********************************************
             * 
             * Debug
            */
            string debugStr = "ctSelMode.calNonCal - Calc Tool Edit/View Tab Received:\n";

            foreach (var boxSel in boxSelBuff)
            {
                debugStr += (boxSel.box + ", ");
            }

            //MessageBox.Show(new Form { TopMost = true }, debugStr, "ctSelMode.calNonCal:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.information, debugStr))
            {
                mainMsg.ShowDialog(this);
            }

            boxSelBuff.Clear();

            /*
             * *********************************************
            */
        }

        private void ProcessCreateStdBoxSel()
        {
            /*
             * We need to check the Calc Tool selection mode to determine what
             * to do with the box(es) held in boxSelBuff.
            */
            switch (selMode)
            {
                case (ctSelMode.none):

                    /*
                     * This should not happen. We will alert user and clear the selection buffer.
                    */
                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.warning, "ProcessCreateStdBoxSel():\n\nCalc Tool received selected box(es) while selection mode is 'none'."))
                    {
                        mainMsg.ShowDialog(this);
                    }

                    boxSelBuff.Clear();

                    // Possibly broadcast an unselect event back to the form...?

                    break;

                case (ctSelMode.crStdCalcSingleNonCal):

                    /*
                     * This is the box that we are creating a calculation for. We use the 'Calc Format'
                     * combo box to set the calc format string if it is not equal to 'NONE'.
                    */
                    string calcBoxAssign = boxSelBuff[0].box + " = ";
                    string calcAttr = createStdFormatComboBox.SelectedItem.ToString();

                    if (calcAttr == "NONE")
                    {
                        calcAttr = "";
                    }
                    else
                    {
                        calcAttr += " % ";
                    }

                    calcCreateStdTextBox.Text = string.Format("{0}{1}()", calcBoxAssign, calcAttr);

                    // Place the cursor between the parens.
                    calcCreateStdTextBox.SelectionStart = calcCreateStdTextBox.Text.Length - 1;

                    /*
                     * Protect the calcBoxAssign from being changed. User will need to
                     * re-start the calculation if they want to change this.
                    */

                    // Now set the selection mode to accept any box - calculated or not.
                    selMode = ctSelMode.crStdCalcCalNonCal;

                    // Set message labels to tell user what to select/do next:
                    createStdCalcMsg1Label.Text = "Now select reference boxes for your calculation, enter values, select operators, etc.";
                    createStdCalcMsg2Label.Text = "The tool will store all selected boxes until you specify a math operator [+, -, *, /].";

                    calcCreateStdTextBox.Focus();

                    boxSelBuff.Clear();

                    break;

                case (ctSelMode.crStdCalcCalNonCal):

                    /*
                     * [11/21/18] We are changing this.....
                     * We are allowing reference box selections, so we just let this buffer
                     * grow until the user clicks an operation button {+, -, *, /}.
                     * 
                     * NEW COMMENT STARTS HERE:
                     * 
                     * We are looking for box selections now that we are 'on the right side of the 
                     * calc equals sign'. We now traverse a simple algorithm in order to take our
                     * next step. Here we need to check the state of the calculation, was the last
                     * insertion into the calc creation textbox a box or a selection of boxes,
                     * say for example, being added together?
                     * 
                     * If so we probably need to tell them the tool can not insert a box or group
                     * of boxes into the calc since it would be box next to box in the string.
                     * 
                     * Else, we will just insert a single box into the calc string at the current
                     * cursor location. If a range of boxes were selected we will need to .show the
                     * DescribeSelections window where we display the box(es) and require that the
                     * user select [describe] which operation we are performing between these boxes.
                     * 
                     * TESTING firing up the DescribeSelections window with these boxes for now:
                     * 
                     * Build the string of selected boxes to pass to the DescribeSelections form.
                    */
                    string selBoxes = "";

                    foreach (var boxSel in boxSelBuff)
                    {
                        selBoxes += (boxSel.box + ", ");
                    }

                    using (var descSel = new DescribeSelections(selBoxes))
                    {
                        /*
                         * Need to grab return to see what mathematical operation they are doing,
                         * or they may be aborting the selection in which case we need to fire
                         * a unselect event back to the parent form.
                         */
                        var result = descSel.ShowDialog(this);

                        switch (result)
                        {
                            case (DialogResult.OK):

                                MessageBox.Show(string.Format("'OK' From DescribeSelections!\nUser has selected {0}", descSel.mathOperator));
                                break;

                            case (DialogResult.Abort):

                                MessageBox.Show("'Abort' From DescribeSelections!");
                                break;

                            case (DialogResult.Cancel):

                                // I don't think this will be seen...It would be windows shutting it down?
                                MessageBox.Show("'Cancel' From DescribeSelections!");
                                break;
                        }
                    }

                    // Don't forget to empty the boxSelBuff and check the return value
                    // for an 'abort selection' indication........

                    /**********************************************
                     * 
                     * Debug
                    */
                    //string debugStr = "\nctSelMode.calNonCal - contents of boxSelBuff: ";

                    //foreach (var boxSel in boxSelBuff)
                    //{
                    //    debugStr += (boxSel.box + ", ");
                    //}

                    //debugStr += "\n";

                    //MessageBox.Show(new Form { TopMost = true }, debugStr, "ctSelMode.calNonCal:", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    /*
                     * *********************************************
                    */

                    break;

                default:

                    /*
                     * This should not happen. We will alert user and clear the selection buffer.
                    */
                    using (MessageForm mainMsg = new MessageForm(MessageForm.msgIcon.warning, "ProcessCreateStdBoxSel():\n\nCalc Tool received selected box(es) but selection mode has not been set."))
                    {
                        mainMsg.ShowDialog(this);
                    }

                    boxSelBuff.Clear();

                    // Possibly broadcast an unselect event back to the form...?

                    break;
            }
        }

        private void ToolBelt_Load(object sender, EventArgs e)
        {
            if (Owner != null)
            {
                // Center CalcTool window in Parent form.
                Location = new Point(Owner.Location.X + Owner.Width / 2 - Width / 2,
                    Owner.Location.Y + Owner.Height / 2 - Height / 2);
            }
        }

        private void toolBeltExitBtn_Click(object sender, EventArgs e)
        {
            /*
             * We un-subscribe from the Form1.CalcToolBoxSel event in 
             * the Tool Belt FormClosed event handler 'ToolBeltClosed'
             * located in Form1.cs.
            */
            this.Close();
        }

        private void toolBeltTabCtrl_Selected(object sender, TabControlEventArgs e)
        {
            // First determine the selected tab page by using the index value.
            switch (toolBeltTabCtrl.SelectedIndex)
            {
                case (0):

                    // Create Standard Calc:
                    TabPageSetup(tbTabPage.boxSel);

                    break;

                case (1):

                    // Create Individually Boxed Calc:
                    TabPageSetup(tbTabPage.codeNav);

                    break;

                case (2):

                    // Edit/View Calc:
                    TabPageSetup(tbTabPage.fcnPick);

                    break;

                case (3):

                    // Replicate Calc:
                    TabPageSetup(tbTabPage.snipPick);
                    break;
            }
        }

        private void createStdCalcStartBtn_Click(object sender, EventArgs e)
        {
            // Testing reading from 'ref DocCalcs dc':
            //calcToolTextBox.Text = docCalcs.calcLinkFile;

            // Testing writing to 'ref DocCalcs dc':
            //docCalcs.calcLinkFile = "Harry was here";

            createStdCalcMsg1Label.Text = "You are now choosing what box to calculate: Select a single non-calculated box.";
            createStdCalcMsg2Label.Text = "Any other box selections will be ignored.";

            /*
             * We are ready to receive a single non-calculated box selection
             * as our box for which we are creating a calculation.
            */
            selMode = ctSelMode.crStdCalcSingleNonCal;

            // Subscribe to the associated event.
            //Form1.CreateStdCalcBoxSel += Form1_CreateStdCalcBoxSel;
        }

        private void crStdCalcPlusBtn_Click(object sender, EventArgs e)
        {
            /*
             * We check the boxSelBuff for any selected boxes. If there are
             * selected boxes we will apply the + operator to them and insert 
             * the addition string into the Calc Tool text box at the cursor
             * location.
             * 
             * If the boxSelBuff is empty we will insert " + " into the
             * text box at the cursor location.
            */
            string addStr = "";
            string boxFormat = "";
            int numBoxes = boxSelBuff.Count;

            if (numBoxes > 0)
            {
                boxFormat = createStdBoxAttrComboBox.SelectedItem.ToString();

                for (int i = 0; i < numBoxes; i++)
                {
                    if ((i < numBoxes - 1) || (numBoxes == 1))
                    {
                        //addStr += (boxSelBuff[i].box + " + ");
                        addStr += (string.Format("{0}{1} + ", boxSelBuff[i].box, boxFormat));
                    }
                    else
                    {
                        //addStr += (boxSelBuff[i].box);
                        addStr += (string.Format("{0}{1}", boxSelBuff[i].box, boxFormat));
                    }
                }

                // Clear the buffer for the next operator:
                boxSelBuff.Clear();
            }
            else
            {
                addStr += " + ";
            }

            // Inserting at the current cursor loation.
            calcCreateStdTextBox.SelectedText = addStr;
        }

        private void ctrlArrayList_Click(object sender, EventArgs e)
        {
            if (ctrlArrayList.SelectedIndex != -1)
            {
                _ctClient.GotoCode(_docFile.ctrlArrays[ctrlArrayList.SelectedIndex]);
            }
        }

        private void docFcnList_Click(object sender, EventArgs e)
        {
            codeNavFcnListIndex = docFcnList.SelectedIndex;

            if (docFcnList.SelectedIndex != -1)
            {
                //MessageBox.Show(new Form { TopMost = true }, "Click/Selection!", "docFcnList_Click():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                _ctClient.GotoCode(_docFile.docFunctions[docFcnList.SelectedIndex]);
            }
        }

        private void comMarkerList_Click(object sender, EventArgs e)
        {
            codeNavComListIndex = comMarkerList.SelectedIndex;

            if (comMarkerList.SelectedIndex != -1)
            {
                //MessageBox.Show(new Form { TopMost = true }, "Click/Selection!", "comMarkerList_Click():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                _ctClient.GotoCode(_docFile.comMarkers[comMarkerList.SelectedIndex]);
            }
        }

        private void orphanBoxList_Click(object sender, EventArgs e)
        {
            codeNavOrphanListIndex = orphanBoxList.SelectedIndex;

            if (orphanBoxList.SelectedIndex != -1)
            {
                //MessageBox.Show(new Form { TopMost = true }, "Click/Selection!", "orphanBoxList_Click():", MessageBoxButtons.OK, MessageBoxIcon.Information);

                _ctClient.GotoCode(_docFile.orphanBoxes[orphanBoxList.SelectedIndex]);
            }
        }

    }
}
